#!/usr/bin/perl -w
#################################################################################
# Prepare data about the given component to be inserted into the database.
#
# This script uses the output of the system package manager (rpm, dpkg or
# emerge), analyzing the files that are provided by the component.
#
# The script creates a text containing list of components, libraries and
# interfaces (actually for each shared object and executable the output of
# 'readelf' is placed into the output file; see README_collect for details).
# If the script is called for a standalone component then the output file should
# be processed by the 'upload_distr_data.pl' script on the server side for actual
# data insertion. In all other cases files produced by componenttodb.pl will
# be joined by apptodb.pl or distrtodb.pl and resultin files will be processed
# by appropriate upload* scripts.
#
# Classes will be detected while parsing interface names in the 'upload_data.pl'
#
# Symlinks and files that are not executable or shared objects are ignored.
#
# For shared objects, the script obtains ABItag using objdump
#
# Author: Denis Silakov <silakov@ispras.ru>
#################################################################################

use strict;
use Getopt::Long;

use Text::Wrap;
Getopt::Long::Configure ("posix_default", "no_ignore_case");

use Distros;

# Component information
my $ComponentName;
my $ComponentVersion="";
my $ComponentId=0;
my $ComponentArch;
my $ComponentDbName;

# Package manager used in the distribution
my $PackageManager="";

# All about rpms
my @rpms;     # rpms that provide the component
my @sofiles;  # list of files from rpms that provide the component

# All about library under process
my $LibName;
my $LibRunName = "";
my $LibSoName = "";
my $FileName = "";
my $LibPath = "";
my $LibPath_p = "";
# Options
my $opt_help;
my $OutputFile;
my $ComponentFilesList;
my $quiet=1;
my $totally_quiet;
my $trsp_symlink;

my %path_subst; # Replacements for temporary filenames

GetOptions("help|h!" => \$opt_help,
            "verbose" => sub {$quiet=0;},
            "quiet|q!" => \$totally_quiet,
            "arch|a=s" => \$ComponentArch,
            "packagemanager|m=s" => \$PackageManager,
            "comp-dbname|c=s" => \$ComponentDbName,
            "version|v=s" => \$ComponentVersion,
            "filelist|l=s" => \$ComponentFilesList,
            "file|f=s" => \$OutputFile,
            "trsp-symlink" => sub {$trsp_symlink=1;},
            );

sub HELP_MESSAGE();
sub add_component();
sub add_library();

# Functions for obtaining information using different package managers
sub get_info_rpm();
sub get_info_dpkg();
sub get_info_emerge();
sub get_info_by_filelist();

HELP_MESSAGE() if (defined $opt_help);

if(@ARGV != 1) {HELP_MESSAGE(); exit(1);}

my $tmpFile = "/var/tmp/componenttodb_lib_$$";

# Guess package manager, if it was not specified
if( (not $PackageManager) and (not $ComponentFilesList) ) {
    $PackageManager = guess_package_manager();
    if( $PackageManager eq "unknown" ) {
        print STDERR "Cannot guess package manager! Please specify it directly using '-p' option.";
        exit(1);
    }
    else {
        print STDERR "No package manager was specified, guessing it is $PackageManager\n";
    }
}

$ComponentName = $ARGV[0];

if( not $ComponentDbName ) {
    $ComponentDbName = $ComponentName;
}

if( !defined $OutputFile ) {
    $OutputFile = $ComponentName."_upload_data";
    $OutputFile =~ s/\//_/g;
}

open( COMPONENT_DATA, "> $OutputFile" );

if( not $ComponentArch ) {
    $ComponentArch = guess_platform(`uname -m`);
}

##############################################################################
# Main section
##############################################################################

if( $ComponentFilesList ) {
    get_info_by_filelist();
}
elsif( $PackageManager eq "rpm" ) {
    get_info_rpm();
}
elsif( $PackageManager eq "dpkg" ) {
    get_info_dpkg();
}
elsif( $PackageManager eq "emerge" ) {
    if( does_shell_know("equery") ) {
        get_info_emerge();
    }
    else {
        die "This script requires 'equery' tool from 'gentoolkit' package";
    }
}
else {
    die "Package manager '$PackageManager' is not supproted";
}

my $python_email_processed=0;
my $python_gtk_processed=0;

# Check the types of every file and call appropriate function
foreach my $soname ( @sofiles )
{
    print "Processing file: '$soname'\n" if(not $quiet);

    # What does 'file' think about our file -
    #  is it an executable or a shared object?
    my $type = `file '$soname'`;

    my $basename = `basename '$soname'`;
    chomp($basename);

    my $comment = "";

    my $soname_p;
    if ( exists $path_subst{$soname} ) {
        $soname_p = $path_subst{$soname};
    } else {
        $soname_p = $soname;
    }

    # Process symlinks
    if( -l($soname) ) {
        my $target_name = "";
        my $target_shortname = "";
        do { # resolve the chain of symlinks
            if( $type =~ /link to `(.+)'/ ) {
                $target_name = $1;

                # Relative path?
                if( $target_name !~ /^\// ) {
                    $target_name = calc_abs_path( $soname, $target_name );
                }

                $target_shortname = $target_name;
                if( $target_name !~ /\// ) {
                    my $target_base = $soname;
                    $target_base =~ s/\Q$basename\E$/\Q$target_name\E/;
                    $target_name = $target_base;
                }
                $type = `file $target_name`;
            }
            else {
                last;
            }
        } while ($type =~ /link to `(.+)'/);

        $target_shortname = `basename $target_name`;
        chomp($target_shortname);

        if( $type !~ /executable/i and !$trsp_symlink and !defined($known_commands{$target_shortname}) ) {
            print "    Symlink to not-executable, skipping...\n" if(not $quiet);
            next;
        }

        if( defined($known_commands{$target_shortname}) ) {
            $type = 'executable';
        }

        $comment = "Symlink to $target_name";
    }

    if( $type =~ m/\:\s*(\w(\s|\:|\-|\/|\w|\+|\.)*)\s+script/ ) {
        if( defined( $known_wrong_commands{$basename} ) ) {
            # Skip files sometimes reported as commands/scripts, but which are known to be not commands or scripts
            next;
        }
        print "Found a script: $soname\n" if(not $quiet);
        print COMPONENT_DATA "!Script '$basename' '$soname_p' '$1'\n";
        my $interpreter = `head -n 1 '$soname'`;
        chomp( $interpreter );
        if( $interpreter !~ /^\#\!/ ) {
            warn "Bad interpreter - '$interpreter'" if( !$quiet );
            $interpreter = "";
        }
        else {
            $interpreter =~ s/^\#\!//;
        }

        if( $type =~ /\:.*perl\s+script/i or $interpreter =~ /perl/ ) {
            print COMPONENT_DATA "  PerlInterpreter: $interpreter\n";
            close( COMPONENT_DATA );
            system "./interpreted\\ languages/analyze_perl.pl \"$soname\" >> \"$OutputFile\"";
            open( COMPONENT_DATA, ">> $OutputFile" );
        }
        elsif( $type =~ /\:.*python\s+script/i or $interpreter =~ /python/ ) {
            print COMPONENT_DATA "  PythonInterpreter: $interpreter\n";
            close( COMPONENT_DATA );
            system "./interpreted\\ languages/analyze_python.py \"$soname\" >> \"$OutputFile\"";
            open( COMPONENT_DATA, ">> $OutputFile" );
        }
        else {
            print COMPONENT_DATA "  ScriptInterpreter: $interpreter\n";
        }

        if( $soname =~ /\.pm$/ and $type =~ /text/ ) {
            print "Found a perl module: $soname\n" if(not $quiet);
            my $modname = $basename;
            $modname =~ s/\.p(m|l)$//i;
            open( PACKAGES, "cat \"$soname\" | LC_ALL=C grep package |" );
            my $found_package_record = 0;
            while( <PACKAGES> ) {
                chomp;
                if( /package\s+((\S|\:)+)\s*\;/ ) {
                    print COMPONENT_DATA "!PerlModuleProvided '$1' '$soname_p'\n";
                    $found_package_record = 1;
                }
            }
            close( PACKAGES );
            if( !$found_package_record ) {
                print COMPONENT_DATA "!PerlModuleProvided '$modname' '$soname_p'\n";
            }
            close( COMPONENT_DATA );
            system "./interpreted\\ languages/analyze_perl.pl \"$soname\" >> \"$OutputFile\"";
            open( COMPONENT_DATA, ">> $OutputFile" );
        }
        elsif( $soname =~ /\.pyc?$/i ) { # ! experimental
            my $pythonResult = `./interpreted\\ languages/analyze_python.py \"$soname\"`;
            if( $pythonResult !~ /Cannot analyse/ ) {

            # See comments below for such checks
#               my $modname = $basename;
#               $modname =~ s/\.pyc?$//i;
#               if( $modname eq "__builtin__" or $modname eq "__main__" or $modname eq "__init__" or $modname eq "__future__" ) {
#                   next;
#               }
#               my $py_import_path = $soname;
#               $py_import_path =~ s/\Q\/$basename\E//;
#               my $errors = "";
#               if( $quiet ) {
#                   $errors = "2>/dev/null";
#               }
#
#               my $exports_str = `PYTHONPATH=\$PYTHONPATH:$py_import_path python -c "import sys; import os; import $modname; print os._get_exports_list($modname)" $errors | sed s/\\\\\\[// | sed s/\\\\\\]//| sed s/\\'//g | sed s/\\ //g`;
#               chomp( $exports_str );
#               my @exports = split( ',', $exports_str );
#               my $count = 0;
#               while( $count <= $#exports ) {
#                   my $expname = $exports[$count];
#                   print COMPONENT_DATA "!PythonModule '$modname.$expname' '$soname_p' 'python'\n";
#                   $count++;
#               }

                print COMPONENT_DATA "!PythonModule '$basename' '$soname_p' 'python'\n";
            }
        }

        next;
    }
    elsif( $type =~ /\:.*perl.+module/i or ($soname =~ /\.pm$/ and $type =~ /text/ ) ) {
        print "Found a perl module: $soname\n" if(not $quiet);
        my $modname = $basename;
        $modname =~ s/\.p(m|l)$//i;
        open( PACKAGES, "cat \"$soname\" | LC_ALL=C grep package |" );
        my $found_package_record = 0;
        while( <PACKAGES> ) {
            chomp;
            if( /package\s+((\S|\:)+)\;$/ ) {
                print COMPONENT_DATA "!PerlModuleProvided '$1' '$soname'\n";
                $found_package_record = 1;
            }
        }
        close( PACKAGES );
        if( !$found_package_record ) {
            print COMPONENT_DATA "!PerlModuleProvided '$modname' '$soname_p'\n";
        }
        close( COMPONENT_DATA );
        system "./interpreted\\ languages/analyze_perl.pl \"$soname\" >> \"$OutputFile\"";
        open( COMPONENT_DATA, ">> $OutputFile" );
        next;
    }
    elsif( $type =~ m/MS-DOS executable/ ) {
        print "Found MS-DOS executable: $soname\n" if(not $quiet);
        print COMPONENT_DATA "!Command '$basename' '$soname_p'\n";
        next;
    }
    elsif( $type =~ m/executable/ ) {
        if( defined( $known_wrong_commands{$basename} ) ) {
            # Skip files sometimes reported as commands, but which are known to be not commands
            next;
        }
        print "Found a command: $soname\n" if(not $quiet);
        if( $soname =~ /\.py$/ ) {
            print COMPONENT_DATA "!PythonModule '$basename' '$soname_p' 'python'\n";
        }
        print COMPONENT_DATA "!Command '$basename' '$soname_p' '$comment'\n";
        if( $comment =~ /Symlink/ and !$trsp_symlink ) {
            next;
        }

        system "readelf -WhlSsdA \"$soname\" > $tmpFile";
        close( COMPONENT_DATA );
        system "cat $tmpFile >> \"$OutputFile\"";
        open( COMPONENT_DATA, ">> $OutputFile" );
        system "rm -f $tmpFile";
        next;
    }
    elsif( $type =~ m/shared object/ ) {
        # Sometimes commands are also reported to be 'shared objects' by file...
        # Let's check our special arrays...
        if( defined( $known_commands{$basename} ) ) {
            print "Found a command: $soname\n" if(not $quiet);
            print COMPONENT_DATA "!Command '$basename' '$soname_p'\n";

            system "readelf -WhlSsdA \"$soname\" > $tmpFile";
            close( COMPONENT_DATA );
            system "cat $tmpFile >> \"$OutputFile\"";
            open( COMPONENT_DATA, ">> $OutputFile" );
            system "rm -f $tmpFile";
            next;
       }

        print "Found a shared object: $soname\n" if(not $quiet);

        $LibRunName = $basename;

        $LibPath = $soname;
        $LibPath =~ s/\/\Q$basename\E//;

        $LibPath_p = $soname_p;
        $LibPath_p =~ s/\/[^\/]*$//;

        $FileName = $soname;
        ($LibName, my @temp) = split /\.so/, $LibRunName;

        # Add library data
        add_library();

        system "readelf -WhlSsdA \"$soname\" > $tmpFile";
        close( COMPONENT_DATA );
        system "cat $tmpFile >> \"$OutputFile\"";
        open( COMPONENT_DATA, ">> $OutputFile" );
        system "rm -f $tmpFile";
    }
    elsif( $soname =~ /\.pyc?$/i ) {
        my $interpreter = `head -n 1 '$soname'`;
        chomp( $interpreter );
        my $pythonResult = `./interpreted\\ languages/analyze_python.py \"$soname\"`;
        if( $pythonResult !~ /Cannot analyse/ ) {
            my $modname = $basename;
            $modname =~ s/\.pyc?$//i;
            if( $modname eq "__builtin__" or $modname eq "__main__" or $modname eq "__init__" or $modname eq "__future__" ) {
                next;
            }

            # !!! Next checks are too rough. First, we'll obtain too many functions;
            #     second, importing modules can be dangerous. So do some primitive checks
            #     for 'os' exports only
            if( $modname eq "os" ) {
                my $exports_str = `python -c "import sys; import os; print os._get_exports_list(os)" 2>/dev/null | sed s/\\\\\\[// | sed s/\\\\\\]//| sed s/\\'//g | sed s/\\ //g`;
                foreach my $submod_required ('altsep', 'curdir', 'pardir', 'sep', 'pathsep', 'linesep', 'defpath', 'name', 'path', 'devnull') {
                    if( $exports_str =~ /$submod_required/ ) {
                        print COMPONENT_DATA "!PythonModule 'os.$submod_required' '$soname_p' 'python'\n";
                    }
                }
            }
            elsif( $soname =~ /email/ and !$python_email_processed ) {
                my $exports_str = `python -c "import sys; import os; import email; print os._get_exports_list(email)" 2>/dev/null | sed s/\\\\\\[// | sed s/\\\\\\]//| sed s/\\'//g | sed s/\\ //g`;
                foreach my $submod_required ('base64MIME', 'Charset', 'Encoders', 'Errors', 'Generator', 'Header',
                                'Iterators', 'Message', 'MIMEAudio', 'MIMEBase', 'MIMEImage', 'MIMEMessage', 'MIMEMultipart',
                                'MIMENonMultipart', 'MIMEText', 'Parser', 'quopriMIME', 'Utils') {
                    if( $exports_str =~ /$submod_required/ ) {
                        print COMPONENT_DATA "!PythonModule 'email.$submod_required' '$soname_p' 'python'\n";
                    }
                }
                $python_email_processed = 1;
            }
            elsif( $soname =~ /\/gtk\// and !$python_gtk_processed ) {
                my $exports_str = `python -c "import sys; import os; import gtk; print os._get_exports_list(gtk)" 2>/dev/null | sed s/\\\\\\[// | sed s/\\\\\\]//| sed s/\\'//g | sed s/\\ //g`;
                foreach my $submod_required ('gdk') {
                    if( $exports_str =~ /$submod_required/ ) {
                        print COMPONENT_DATA "!PythonModule 'gtk.$submod_required' '$soname_p' 'python'\n";
                    }
                }
                $python_gtk_processed = 1;
            }

            # Try to detect what does the module provide, since some of its exports might be
            #  required by apps directly
#           my $py_import_path = $soname;
#           $py_import_path =~ s/\Q\/$basename\E//;
#
#           my $errors = "";
#           if( $quiet ) {
#               $errors = "2>/dev/null";
#           }
#
#           my $exports_str = `PYTHONPATH=\$PYTHONPATH:$py_import_path python -c "import sys; import os; import $modname; print os._get_exports_list($modname)" $errors | sed s/\\\\\\[// | sed s/\\\\\\]//| sed s/\\'//g | sed s/\\ //g`;
#           chomp( $exports_str );
# #             print "$soname exploited to $exports_str\n";
#           my @exports = split( ',', $exports_str );
#           my $count = 0;
#           while( $count <= $#exports ) {
#               my $expname = $exports[$count];
#               print COMPONENT_DATA "!PythonModule '$modname.$expname' '$soname_p' 'python'\n";
#               $count++;
#           }
#
            if( $interpreter =~ /^\#\!/ ) {
                $interpreter =~ s/^\#\!//;
                print COMPONENT_DATA "!Script '$basename' '$soname_p' 'python'\n";
            }
            else {
                $interpreter = "";
                print COMPONENT_DATA "!PythonModule '$basename' '$soname_p' 'python'\n";
            }
            print COMPONENT_DATA "  PythonInterpreter: $interpreter\n";
            print COMPONENT_DATA $pythonResult;
        }
    }
    else {
        print "    Not executable or shared object, not a perl/python script/object, skipping...\n" if( not $quiet );
    }
}

################################################################################################
# Collect information about components directly by file list or using different package managers
################################################################################################

sub get_info_by_filelist()
{
    open( FILE_LIST, "< $ComponentFilesList" ) || die "Cannot open file with filelist ($ComponentFilesList)";

    my $tmpName = $ComponentName;
    if( $ComponentName =~ /rpm$/ ) {
        $ComponentName =~ s/\/.*\///g;
        $ComponentName =~ s/\-[0-9].*//g;
        $ComponentDbName = $ComponentName;
    }
    elsif( $ComponentName =~ /deb$/ ) {
        $ComponentName =~ s/\/.*\///g;
        $ComponentName =~ s/\_.*//g;
        $ComponentDbName = $ComponentName;
    }
    else {
        $ComponentName = $ComponentDbName;
    }
    add_component();
    $ComponentName = $tmpName;

    my $type;

    while( my $filename = <FILE_LIST> ) {
        chomp($filename);

        my $subst = undef;
        # There could be specified a replacement for path/filename
        # that should be printed in the output file.
        # (usually to replace temp names with smth more useful).
        # It goes after first '|'
        if ( $filename =~ /^([^\|]*)\|(.*)$/ ) { # look for substitution
            $subst = $filename;
            $filename = $1;
        }
        $type = `file '$filename'`;

        if( $type =~ /directory/ ) {
            local *find_files = sub {
                my ($cmd) = @_;

                open( FILES, "$cmd |" )
                    or die "Cannot process files from directory $_";

                while( my $found = <FILES> ) {
                    chomp $found;
                    if ( defined $subst ) {
                        my $new_subst = $found;
                        $new_subst =~ s/^\Q$filename\E/$subst/;
                        $path_subst{$found} = $new_subst;
                    }
                    push @sofiles, $found;
                }
                close FILES;
            };

            &find_files( "find '$filename' -type f -print" );

            # Process symlinks, too
            &find_files( "find '$filename' -type l -print" );
        }
        else {
            push @sofiles, $filename;
            if ( defined $subst ) {
                $path_subst{$filename} = $subst;
            }
        }
    }

    close FILE_LIST;
}

sub get_info_rpm()
{
    # Get a list of rpm that provide the given component
    open(RPM_PROV,"rpm -q --whatprovides $ComponentName|") || die "Cannot get information from rpm.";

    @rpms = <RPM_PROV>;

    chomp($rpms[0]);
    if( $rpms[0] eq "no package provides $ComponentName" ) {
        # Maybe we are processing an application and 'whatprovides' doesn't
        # know about it?
        # (If we are wrong here and $ComponentName is not a valid rpm name,
        #  we'll not obtain the file list - nothing wrong, but user should be informed)
        $rpms[0] = $ComponentName;
    }

    close RPM_PROV;

    # If component version is not set manually, then take it from the first rpm
    if( !$ComponentVersion ) {
        if ($rpms[0] =~ /((\w|\d|\-\w)*)\-((\w|\d|\.|\-)*)/ ) {
            $ComponentVersion = $3;
        }
        else {
            print STDERR "$ComponentName: Cannot get component version!\n";
            $ComponentVersion = "";
        }
    }

    add_component();

    # Explore files provided by rmps
    for( 0..$#rpms ) {
        chomp($rpms[$_]);
        my $request = '--fileprovide';
        # old rpm uses 'filesbypkg' instead of 'fileprovide'
        my $try = `rpm -q --fileprovide $rpms[$_]`;
        if( $try =~ /unknown\s*option/ ) {
            $try = `rpm -q --filesbypkg $rpms[$_]`;
            if( $try =~ /unknown\s*option/ ) {
                die "Cannot extract app file list from rpm!";
            }
            else {
                $request = '--filesbypkg';
            }
        }

        open(FILE_PROV,"rpm -q $request ".$rpms[$_]." | ") || die "Cannot get information from rpm.";

        while(<FILE_PROV>) {
            chomp;

            if( /not installed / ) {
                warn "Missing package!";
                next;
            }

            if( /((\/(\w|\.|\-|\+|\[)+)*)\s*((\w+\/\s*)+)*/ ) {
                print "file found: '".$1."'\n" if(not $quiet);
                push @sofiles, $1;
            }
        }

        close FILE_PROV;
    }
}

sub get_info_dpkg()
{
    # Unfortunately, we cannot use something like 'rpm -q --whatprovides' here,
    # cause for many base components (like Qt, Gtk etc.) there are no
    # packages that 'Provide' them. So, user should provide us with correct
    # package name

    # Check if the component is available
    my $notAvailable = `dpkg -s "$ComponentName" | grep "is not installed"`;
    if( $notAvailable ) {
        warn "Package $ComponentName is not installed and cannot be processed.";
        return;
    }

    # If component version is not set manually, then take it from the first package
    if( !$ComponentVersion ) {
        my $ver = `dpkg -s "$ComponentName" | LC_ALL=C grep -i version`;
        if( $ver =~ /(V|v)ersion\s*:\s*(\S+)/ ) {
            $ComponentVersion = $2;
        }
        else {
            warn "$ComponentName: Cannot get component version!";
            $ComponentVersion = "";
        }
    }

    add_component();

    open(FILE_PROV,"dpkg -L $ComponentName | ") || die "Cannot get information from dpkg.";
    while(<FILE_PROV>) {
        chomp;
        print "file found: '".$_."'\n" if(not $quiet);
        push @sofiles, $_;
    }

    close FILE_PROV;
}

sub get_info_emerge()
{
    # Get component version, if it is not specified directly

    if( !$ComponentVersion ) {
        my $VerLine = `equery which "$ComponentName"`;
        if( $VerLine =~ /\Q$ComponentName\E-(.+)\.ebuild/ ) {
            $ComponentVersion=$1;
        }
        else {
            warn "Cannot get component version for '$ComponentName'!";
        }
    }

    add_component();

    open(FILE_PROV,"equery files $ComponentName | ") || die "Cannot get information from equery.";
    while(<FILE_PROV>) {
        chomp;
        print "file found: '".$_."'\n" if(not $quiet);
        push @sofiles, $_;
    }

    close FILE_PROV;
}

##############################################################################
# Functions for printing additional information about artifacts
##############################################################################

# Utility function for processing of objdump output.
sub parse_dword {
    my ( $dword, $byte_order ) = @_;

    if ( $byte_order == 0 ) { # little-endian
        my $reversed = "";
        while ( $dword =~ s/^(\d{2})// ) {
            $reversed = $1.$reversed;
        }
        $dword = $reversed;
    }

    return hex($dword);
}

#
# add_library: print record with library information
#
sub add_library()
{
    # Get soname of the library
    $LibSoName='';
    my $SonameString='';
    $SonameString = `readelf -d \"$LibPath/$LibRunName\" | LC_ALL=C fgrep SONAME`;
    if( $SonameString =~ /Library soname: \[((\S)*)\]/ ) {
        $LibSoName = $1;
    }

    # Get ABI-tag
    open OBJDUMP, "objdump -s -j .note.ABI-tag '$LibPath/$LibRunName' |"
            or die "Cannot get information from objdump.";

    my @dwords = (); # Contents of '.note.ABI-tag' section
                     # in 32 bit dwords in a hexadecimal representation
                     # with a platform-specific byte order.

    while ( my $line = <OBJDUMP> ) {
        chomp $line;

        if ( $line =~ /^ \d+/ ) { # Contents of the section .note.ABI-tag
            #F ormat:  offset dword_a{8} dword_b{8} dword_b{8} dword_d{8}  ...ASCII...
            if ( $line =~ /^ [\dA-F]+\s+([\dA-F]{8})\s+([\dA-F]{8})\s+([\dA-F]{8})\s+([\dA-F]{8})\s+/i ) {
                push @dwords, $1, $2, $3, $4;
            }
            else {
                die "Can't parse objdump output: '$line'";
            }
        }
    }

    my $ABItag = "";

    # Process the obtained data
    if ( @dwords ) {
        scalar(@dwords) >= 8 or die "Too little bytes in the section.";

        # See LSB-Core-generic/noteabitag.html

        my $byte_order = 0; # little-endian
        # Guess the byte order. Not the best way, but just works.
        if ( $dwords[2] eq "00000001" ) {
            $byte_order = 1; # big endian
        }

        parse_dword( $dwords[0], $byte_order ) == 4
            or die "The first dword ('namesz' field?) should equals 4";
        parse_dword( $dwords[1], $byte_order ) == 16
            or die "The second dword ('descsz' field) should equals 16";
        parse_dword( $dwords[2], $byte_order ) == 1
            or die "The third dword ('type' field) should equals 1";
        $dwords[3] eq "474e5500"
            or die "The fourth dword ('name' field) should be \"GNU\\0\"";
        parse_dword( $dwords[4], $byte_order ) == 0
            or die "The fifth dword (1st dword in 'desc' field) should be 0 (this signifies a Linux executable)";

        my $a = parse_dword( $dwords[5], $byte_order );
        my $b = parse_dword( $dwords[6], $byte_order );
        my $c = parse_dword( $dwords[7], $byte_order );

        $ABItag = "Linux $a.$b.$c";
    }

    # Not perfect, but we need some way to distinct python modules provided as *.so files
    if( $LibPath =~ /python/ ) {
        print COMPONENT_DATA "!PythonModule '$LibName' '$LibPath_p/$LibName' 'python'\n";
    }

    # Try to detect library version (quite difficult task...)
    if( ($LibName =~ /^ISO/) or ($LibName =~ /^GOST/) or ($LibName =~ /^UTF/) or ($LibName =~ /^TIS/)
            or ($LibName =~ /^ISIRI/) ) {
        print COMPONENT_DATA "!Library '$LibName' '' '$LibRunName' '$LibPath_p' '$ABItag' '$LibSoName'\n";
    }
    elsif( $LibName =~ /(lib(g|stdc)\+\+\-(\d-)?libc)(\d\..+)/ ) {
        my $lname = $1;
        my $lver = $4;
        print COMPONENT_DATA "!Library '$lname' '$lver' '$LibRunName' '$LibPath_p' '$ABItag' '$LibSoName'\n";
    }
    elsif( $LibName =~ /((\w|\d|_|\+|\-\w)+)\-((\d).+)/ ) {
        my $lname = $1;
        my $lver = $3;
        print COMPONENT_DATA "!Library '$lname' '$lver' '$LibRunName' '$LibPath_p' '$ABItag' '$LibSoName'\n";
    }
    elsif( $LibRunName =~ /((\w|\d|_|\+|\.|\-\w)+)\.so\.((\d).+)/ ) {
        my $lname = $1;
        my $lver = $3;
        print COMPONENT_DATA "!Library '$lname' '$lver' '$LibRunName' '$LibPath_p' '$ABItag' '$LibSoName'\n";
    }
    else {
        print COMPONENT_DATA "!Library '$LibName' '' '$LibRunName' '$LibPath_p' '$ABItag' '$LibSoName'\n";
    }
}

sub add_component()
{
    # Remove the leading '=', if presents, in case of Gentoo portage
    if( $PackageManager eq "emerge" ) {
        my $PackageName = $ComponentName;
        $PackageName =~ s/^\=//;
        if( $PackageName =~ /(-$ComponentVersion)$/ ) {
                $PackageName =~ s/-$ComponentVersion//;
        }
        print COMPONENT_DATA "!Component '$PackageName' '$ComponentVersion' '$ComponentDbName' '$ComponentArch'\n";
    }
    else {
        print COMPONENT_DATA "!Component '$ComponentName' '$ComponentVersion' '$ComponentDbName' '$ComponentArch'\n";
    }
}

##############################################################################
# Miscellaneous functions
##############################################################################

sub HELP_MESSAGE()
{
print STDERR <<"EOM"

$0: Obtain information about libraries, commands and interfaces provided by the given component.

Usage: $0 [options] component_name

component_name                Name of the component to process as it should be
                              given to the package manager

-c|--comp-dbname              Name of the component that should be placed into the database
                              (by default <component_name> is used)

-h|--help                     Display this help message

-f|--file filename            Filename where the data for upload will be situated
                              By default, "<component_name>_upload_data" is used

-l|--filelist filename        File containing list of file names that are installed by the application.
                              (Directory names are also allowed; all files from given directories will be processed)

-v|--version version          Component version. By default, the script will try to obtain version using
                              package manager, but that version might be not that one you'd like to see if
                              the component is provided by several packages, since each package may have
                              its own version

-a|--arch                     Architecture name where the component is processed
                              (if it is not set, the output of 'uname' will be used)

-m|--packagemanager pkg_name  Package manager used in the distribution. Supported values
                              are 'rpm', 'dpkg' and 'emerge'. If none is specified, the script
                              will try to guess it.

-q|--quiet                    Be really quiet - do not produce any messages except critical errors

--verbose                     Be verbose

EOM
      ; exit(1);
}
